package com.ibm.spss.cads.dataservice.adv;

import java.io.File;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Scanner;

import org.apache.log4j.Logger;

import com.spss.adv.dataService.AnalyticDataServiceException;
import com.spss.adv.dataService.IDataServiceRowIterator;
import com.spss.adv.internal.dataService.DataServiceText;
import com.spss.adv.dataService.IDataAdapter;
import com.spss.data.service.internal.datasource.dataservice.DataServiceDataSource;
import com.spss.data.service.multipleinput.IDataRow;
import com.spss.data.service.multipleinput.IDataRowIterator;
import com.spss.data.service.value.Value;
import com.spss.pev.schema.types.PevDataType;
import com.spss.scoring.ScoringMessage;
import com.spss.scoring.provider.record.ValueException;
import com.spss.scoring.provider.record.ValueHelper;

public class SimplifiedADVAdapter implements IDataAdapter{
	
	private String itsDataSourceName = null ;
	private final String prop_data_location = "DataLocation";
	private List<String> prop_keys = null;
	private Map<String, String> dic_map = new HashMap<String, String>();
 
	
	public SimplifiedADVAdapter () {
		String aMethodPrefix = "SimplifiedADVAdapter.init() <> | " ;
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " entered...") ;
		prop_keys = new ArrayList<String>();
		prop_keys.add(prop_data_location);
		prop_keys.add("property.name");
	}
	
	public SimplifiedADVAdapter (DataServiceDataSource theDataSource) {
				
		String aMethodPrefix = "SimplifiedADVAdapter.init <theDataSource> | " ;
		
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " entered...") ;
		
		itsDataSourceName = theDataSource.getName() ;
		
		//initialize map for properites
		prop_keys = new ArrayList<String>();
		prop_keys.add("DataLocation");
		prop_keys.add("property.name");
		
		
		
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + "Data Source Name: " + itsDataSourceName) ;
		
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " exit...") ;
	}

	
		
	/**
	 * The execute method is called by the ADV to acquire the data
	 */
	@Override
	public synchronized IDataServiceRowIterator execute(Collection<String> theFields, List<String> theKeyFields, IDataRowIterator theKeyFieldValues, Map<String, IDataRowIterator> theContextData) throws AnalyticDataServiceException  {

		String aMethodPrefix = "SimplifiedADVAdapter.execute | " ;
		
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " entered...") ;
		
		try {
			
			//attempt to read the properties for the Custom Data Source
			readProperties(theContextData) ;
			
			//try to read csv file
			String FileLocation = dic_map.get("DataLocation");

			String[] FieldsArray = theFields.toArray(new String[0]);
			List<List<String>> records = new ArrayList<>();
			try (Scanner scanner = new Scanner(new File(FileLocation));) {
			    while (scanner.hasNextLine()) {
			        records.add(getRecordFromLine(scanner.nextLine()));
			    }
			}
			catch(Exception ex)
			{
                throw new AnalyticDataServiceException(new ScoringMessage(ex.getMessage()), ex);			
			}

			//To filter the data row with only the keys
			Map<String, String> row_map = new HashMap<String, String>();
			List<HashMap<String, String>> FilteredResults = new ArrayList();
            for(List<String> oneRow : records)
            {
            	row_map.clear();
            	int i=0;
            	for(String oneValue : oneRow)
            	{  
            		row_map.put(FieldsArray[i], oneValue);
            		i++;
            	}
            	
            	String[] KeyFieldsArray = theKeyFields.toArray(new String[0]);
              	String oneField=KeyFieldsArray[0];
            	String valueToCheck = row_map.get(oneField);
            	Boolean found = false;
            	for(IDataRow oneValue : theKeyFieldValues)
            	{
            		String strToBeChecked = oneValue.getValue(oneField).stringValue();
        			if(valueToCheck.compareTo(strToBeChecked)==0)
            		{
            			found = true;
            			HashMap<String, String> mapOneRow = new HashMap<String,String>();
            			mapOneRow.putAll(row_map);
            			FilteredResults.add(mapOneRow);
            	    }
            	}
            	
            	theKeyFieldValues.reset();
            	
            	if(found == false)
            	{
            		row_map.clear();
            	}
            }
            	     
            	
            //To capuslate into DataMapRowIterator
			Result rs = new Result();
			for(HashMap<String,String> oneRow : FilteredResults) {
				ArrayList<Field> rsFields=new ArrayList<Field>();
                for (HashMap.Entry<String, String> entry : oneRow.entrySet()) {
        		     Field oneField = new Field(entry.getKey(), PevDataType.STRING, entry.getValue());
           		     rsFields.add(oneField);
        	    }
                Row row = new Row(rsFields);
                rs.addRow(row);
			}
            
            ResultSetRowIterator rsi = new ResultSetRowIterator(rs);
            return rsi ;
		
		} catch (SQLException sqle) {
			// TODO Auto-generated catch block
			throw new AnalyticDataServiceException(new ScoringMessage(DataServiceText.ATTRIBUTE_DEPENDENCY),sqle);
		} finally {
			Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " exit...") ;
		}
	}
	
	/**
	 * Parsing csv files
	 * 
	 * @param one line from csv file
	 * @returns one row (list) with columns (String)
	 */
	private List<String> getRecordFromLine(String line) {
	    List<String> values = new ArrayList<String>();
	    try (Scanner rowScanner = new Scanner(line)) {
	        rowScanner.useDelimiter(",");
	        while (rowScanner.hasNext()) {
	            values.add(rowScanner.next());
	        }
	    }
	    return values;
	}
	
	/**
	 * Read the properties from the Custom Data Source
	 * 
	 * @param theContextData
	 * @throws AnalyticDataServiceException
	 */
	private synchronized void  readProperties (Map<String, IDataRowIterator> theContextData) throws AnalyticDataServiceException {
		
		String aMethodPrefix = "SimplifiedADVAdapter.readProperties | " ;
		
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " entered...") ;
		
		/**
		 * First collect the properties
		 */
		if (theContextData == null || theContextData.size() == 0) {
			Exception anException = new Exception(aMethodPrefix + "Provided Context is empty") ;
			throw new AnalyticDataServiceException(new ScoringMessage(DataServiceText.ATTRIBUTE_DEPENDENCY),anException);
		}
		
		if (! theContextData.containsKey("property")) {
			Exception anException = new Exception(aMethodPrefix + "Context does not contain property \"property\"") ;
			throw new AnalyticDataServiceException(new ScoringMessage(DataServiceText.ATTRIBUTE_DEPENDENCY),anException);
		}

		IDataRowIterator aContextDataRowIterator = theContextData.get("property") ;
		
		if (aContextDataRowIterator == null || (! aContextDataRowIterator.hasNext())) {
			
			Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + "Attempting to reset the iterator...") ;
			
			//Attempts to reset the object to make sure it starts and the beginning of the iteration
			aContextDataRowIterator.reset() ;
			

			if (aContextDataRowIterator == null || (! aContextDataRowIterator.hasNext())) {
			Exception anException = new Exception(aMethodPrefix + "Context does contain property \"property\", however it is empty.") ;
			throw new AnalyticDataServiceException(new ScoringMessage(DataServiceText.ATTRIBUTE_DEPENDENCY),anException);
			}
		}

		while (aContextDataRowIterator.hasNext()) {
			
			IDataRow aContextDataRow = aContextDataRowIterator.next() ;
			
			if (aContextDataRow == null)
				continue ;

			for (String oneProp : prop_keys) {
			
			      Value aValue1 = aContextDataRow.getValue(oneProp) ;
			
			      if (aValue1 != null) {
				     String aValue = aValue1.stringValue() ;
				
				     Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + oneProp +"set to '" + aValue + "'") ;
				
				     dic_map.put(oneProp, aValue);
			     } 
			 }
		}
		
		Logger.getLogger(SimplifiedADVAdapter.class).info(aMethodPrefix + " exit...") ;
	}
	
	@Override
	public void release() {
		
	}
	
	private final class ResultSetRowIterator implements IDataServiceRowIterator {
	    
		private final Result				itsResult ;
		private final List<IDataRow> 			itsResultData = new ArrayList<IDataRow>();
	    private int								itsPosition = 0 ;
	    private boolean							itsIsDataLoaded = false ;
	    
	    /**
	     * 
	     * @param theFields
	     * @param theResultSet
	     * @throws SQLException
	     */
	    ResultSetRowIterator(Result theResult) throws SQLException, AnalyticDataServiceException {

	    	itsResult = theResult ;
	    }
	    
	    /**
	     * 
	     */
	    public void release() { 
	    }
	    
	    /**
	     * 
	     */
	    public void reset() {
	    	itsPosition = 0;
	    }
	    
	    /**
	     * 
	     */
	    public boolean hasNext() {
	    	
	    	if (! itsIsDataLoaded) {
	    		
	    		for (int aCounter=0;aCounter<itsResult.size();aCounter++) {
	    			IDataRow aDataRow = getDataRow(aCounter);

	    			if (aDataRow == null) 
			          continue ;
			          
			        itsResultData.add(aDataRow) ; 
	    		}
	    		
	    		itsIsDataLoaded = true ;
	    	}
	    	
	    	return itsPosition < itsResultData.size() ;
	    }
	    
	    /**
	     * 
	     */
	    public IDataRow next() {
	    	
	    	if (hasNext()) {
	    		itsPosition = itsPosition + 1 ;
	    		return (IDataRow) itsResultData.get((itsPosition - 1)) ;
	    	} else {
	    		throw new NoSuchElementException() ;
	    	}
	    }
	    
	    /**
	     * getDataRow
	     * 
	     * @return
	     */
	    private IDataRow getDataRow(int theRowRecordCounter) {
	    	

	    	final HashMap<String, Value> aData = new HashMap<String, Value>();
	      
	    	Row aRow = itsResult.getRow(theRowRecordCounter) ;
	    	
	    	if (aRow == null)
	    		return null ;
	    	
	    	
	    	
	    	ArrayList<Field> aFields = aRow.getRow() ;
	    	
	    	for (int aCounter=0;aCounter<aFields.size();aCounter++) {

	    		Field aField = aFields.get(aCounter) ;
	    		
	    		if (aField == null)
	    			continue ;
	    		
    			String aFieldName = aField.getName() ;
    			
    			if (aFieldName == null || aFieldName.length() == 0)
    				continue ;
    			
    			PevDataType aPevDataType = aField.getType() ;
    			
    			if (aPevDataType == null)
    				continue ;
    			
    			String aFielValue = aField.getValue() ;
    			try {
    				aData.put(aFieldName, ValueHelper.getValue(aPevDataType, aFielValue));
    			} catch (ValueException ve) {
 
    			}
	    	}
	    	
	    	IDataRow aResult = new IDataRow() {
	      
	    		public Value getValue(String theField) {
	    			return (Value)aData.get(theField);
	    		}
	    	};
	    	
	    	return aResult ;
	    }
	    
	    public void remove() {
	      throw new UnsupportedOperationException();
	    }
	    
	    public Iterator<IDataRow> iterator() {
	      return this;
	    }
	}
}
